"""Main class."""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

import os
import logging
import re

from typing import Any, Callable, Dict, List, Tuple, Union

from filefinder.matcher import (Matcher, Matches,
                                get_matchers_indices)

log = logging.getLogger(__name__)


class Finder():
    """Find files using a pre-regex.

    Parameters
    ----------
    root : str
        The root directory of a filetree where all files can be found.
    pregex: str
        The pre-regex. A regular expression with added 'Matchers'.
        Only the matchers vary from file to file. See documentation
        for details.
    use_regex: bool
        Characters outside of matcher are considered as valid regex (and not
        escaped).
    replacements : str, optional
        Matchers to replace by a string:
        `'matcher name' = 'replacement string'`.

    Attributes
    ----------
    root: str
        The root directory of the finder.
    pregex: str
        Pre-regex.
    regex: str
        Regex obtained from the pre-regex.
    use_regex: bool
        Characters outside of matcher are considered as valid regex (and not
        escaped) if true.
    pattern: re.pattern
        Compiled regex.
    matchers: list of :class:`Matcher<filefinder.matcher.Matcher>`
        List of matchers for this finder, in order.
    segments: list of str
        Segments of the pre-regex. Used to replace specific matchers.
        `['text before matcher 1', 'matcher 1',
        'text before matcher 2, 'matcher 2', ...]`
    fixed_matchers: dict
        Dictionnary of matchers with a set value.
        'matcher index': 'replacement string'
    _files: list of tuples
        List of tuples containing the filename relative to the root,
        and :class:`Matches<filefinder.matcher.Matches>` object.
    scanned: bool
        If the finder has scanned files.

    """

    def __init__(self, root: str, pregex: str, use_regex: bool = False,
                 **replacements: str):

        if isinstance(root, (list, tuple)):
            root = os.path.join(*root)
        self.root = root

        self.pregex = ''
        self.regex = ''
        self.use_regex = use_regex
        self.pattern = None
        self.matchers = []
        self.segments = []
        self.fixed_matchers = dict()
        self._files = []
        self.scanned = False

        self.set_pregex(pregex, **replacements)
        self.scan_pregex()
        self.update_regex()

    @property
    def n_matchers(self) -> int:
        """Number of matchers in pre-regex."""
        return len(self.matchers)

    @property
    def files(self) -> List[Tuple[str, Matches]]:
        """List of filenames and their matches.

        Will scan files if not already.
        """
        if not self.scanned:
            self.find_files()
        return self._files

    def __repr__(self):
        return '\n'.join([super().__repr__(), self.__str__()])

    def __str__(self):
        s = ['root: {}'.format(self.root)]
        s += ["pre-regex: {}".format(self.pregex)]
        if self.regex is not None:
            s += ["regex: {}".format(self.regex)]
        else:
            s += ["regex not created"]
        if self.fixed_matchers:
            s += ["fixed matchers:"]
            s += ["\t fixed #{} to {}".format(i, v)
                  for i, v in self.fixed_matchers.items()]
        if not self.scanned:
            s += ["not scanned"]
        else:
            s += ["scanned: found {} files".format(len(self._files))]
        return '\n'.join(s)

    def get_files(self, relative: bool = False,
                  nested: List[str] = None) -> List[str]:
        """Return files that matches the regex.

        Lazily scan files: if files were already scanned, just return
        the stored list of files.
        Scanned files are flushed if the regex is changed (by fixing matcher
        for instance).

        Parameters
        ----------
        relative : bool
            If True, filenames are returned relative to the finder
            root directory. If not, paths are absolute. Defaults to False.
        nested : list of str
            If not None, return nested list of filenames with each level
            corresponding to a group in this argument. Last group in the list
            is at the innermost level. A level specified as None refer to
            matchers without a group.

        Raises
        ------
        KeyError: A level in `nested` is not in the pre-regex groups.
        """
        def make_abs(f):
            return os.path.join(self.root, f)

        def get_match(m, group):
            return ''.join([m_.get_match(parsed=False) for m_ in m
                            if m_.matcher.group == group])

        def nest(files_matches, groups, relative):
            if len(groups) == 0:
                return [make_abs(f) if not relative else f
                        for f, m in files_matches]

            group = groups[0]
            files_grouped = []
            matches = {}
            for f, m in files_matches:
                match = get_match(m, group)
                if match not in matches:
                    matches[match] = len(matches)
                    files_grouped.append([])
                files_grouped[matches[match]].append((f, m))

            return [nest(grp, groups[1:], relative) for grp in files_grouped]

        if not self.scanned:
            self.find_files()

        if nested is None:
            files = [make_abs(f) if not relative else f
                     for f, m in self._files]
        else:
            groups = [m.group for m in self.matchers]
            for g in nested:
                if g not in groups:
                    raise KeyError(f'{g} is not in Finder groups.')
            files = nest(self._files, nested, relative)

        return files

    def fix_matcher(self, key: Union[int, str], value: Union[Any, List[Any]],
                    fix_discard: bool = False):
        """Fix a matcher to a string.

        Parameters
        ----------
        key : int, or str, or tuple of str of length 2.
            If int, is matcher index, starts at 0.
            If str, can be matcher name, or a group and name combination with
            the syntax 'group:name'.
            When using strings, if multiple matchers are found with the same
            name or group/name combination, all are fixed to the same value.
        value : str or value, or list of
            Will replace the match for all files. Can be a string, or a value
            that will be formatted using the matcher format string.
            A list of values will be joined by the regex '|' OR.
            Special characters should be properly escaped in strings.
        fix_discard: bool
            If true, matchers with the 'discard' option will still be fixed.
            Defaults to false.
        """
        for m in self.get_matchers(key):
            if not fix_discard and m.discard:
                continue
            self.fixed_matchers[m.idx] = value
        self.update_regex()

    def fix_matchers(self, fixes: Dict[Union[int, str], Any] = None,
                     fix_discard: bool = False,
                     **fixes_kw):
        """Fix multiple values at once.

        Parameters
        ----------
        fixes: dict
           Dictionnary of matcher key: value. See :func:`fix_matcher` for
           details. If None, no matcher will be fixed.
        fix_discard: bool
            If true, matchers with the 'discard' option will still be fixed.
            Defaults to false.
        """
        if fixes is None:
            fixes = {}
        fixes.update(fixes_kw)
        for f in fixes.items():
            self.fix_matcher(*f, fix_discard=fix_discard)

    def unfix_matchers(self, *keys: str):
        """Unfix matchers.

        Parameters
        ----------
        keys: str
           Keys to find matchers to unfix. See :func:`get_matchers`.
           If no key is provided, all matchers will be unfixed.
        """
        if not keys:
            self.fixed_matchers = {}
        else:
            for key in keys:
                matchers = self.get_matchers(key)
                for m in matchers:
                    self.fixed_matchers.pop(m.idx, None)
        self.update_regex()

    def get_matches(self, filename: str,
                    relative: bool = True) -> Matches:
        """Get matches for a given filename.

        Apply regex to `filename` and return a dictionary of the results.

        Parameters
        ----------
        filename: str
            Filename to retrieve matches from.
        relative: bool
            Is true if the filename is relative to the finder root directory.
            If false, the filename is made relative before being matched.
            Default to true.

        Raises
        ------
        AttributeError: The regex is empty.
        ValueError: The filename did not match the pattern.
        IndexError: Not as many matches as matchers.
        """
        if not self.regex:
            raise AttributeError("Finder is missing a regex.")

        if not relative:
            filename = os.path.relpath(filename, self.root)

        return Matches(self.matchers, filename, self.pattern)

    def get_filename(self, fixes: Dict = None, relative: bool = False,
                     **kw_fixes: Any) -> str:
        """Return a filename.

        Replace matchers with provided values.
        All matchers must be fixed prior, or with `fixes` argument.

        Parameters
        ----------
        fixes: dict
            Dictionnary of fixes (matcher key: value). For details, see
            :func:`fix_matcher`. Will (temporarily) supplant matcher fixed
            prior. If prior fix is a list, first item will be used.
        relative: bool
            If the filename should be relative to the finder root directory.
            Defaults to False.
        kw_fixes:
            Same as fixes. Takes precedence.

        Raises
        ------
        ValueError: use_regex is activated.
        """
        if self.use_regex:
            raise ValueError("Cannot generate a valid filename if regex "
                             "is present outside matchers.")

        fixed_matchers = self.fixed_matchers.copy()
        if fixes is None:
            fixes = {}
        fixes.update(kw_fixes)
        for key, value in fixes.items():
            for m in self.get_matchers(key):
                fixed_matchers[m.idx] = value

        non_fixed = [i for i in range(self.n_matchers)
                     if i not in fixed_matchers]
        if any(non_fixed):
            log.error("Matchers not fixed: %s",
                      ', '.join([str(self.matchers[i]) for i in non_fixed]))
            raise TypeError("Not all matchers were fixed.")

        segments = self.segments.copy()

        for idx, value in fixed_matchers.items():
            if isinstance(value, (list, tuple)):
                value = value[0]
            if not isinstance(value, str):
                value = self.matchers[idx].format(value)
            segments[2*idx+1] = value

        filename = ''.join(segments)

        if not relative:
            filename = os.path.join(self.root, filename)

        return filename

    def get_func_process_filename(self, func: Callable, relative: bool = True,
                                  *args, **kwargs) -> Callable:
        r"""Get a function that can preprocess a dataset.

        Written to be used as the 'process' argument of
        `xarray.open_mfdataset`. Allows to use a function with additional
        arguments, that can retrieve information from the filename.

        Parameters
        ----------
        func: Callable
            Input arguments (`xarray.Dataset`, filename: `str`,
            `Finder`, \*args, \*\*kwargs)
            Should return a Dataset.
            Filename is retrieved from the dataset encoding attribute.
        relative: If True, `filename` is made relative to finder root.
            This is necessary to match the filename against the finder regex.
            Defaults to True.
        args: optional
            Passed to `func` when called.
        kwargs: optional
            Passed to `func` when called.

        Returns
        -------
        Callable
             Function with the signature of the 'process' argument of
             `xarray.open_mfdataset`.

        Examples
        --------
        This retrieve the date from the filename, and add a time dimensions
        to the dataset with the corresponding value.
        >>> from filefinder import library
        ... def process(ds, filename, finder, default_date=None):
        ...     matches = finder.get_matches(filename)
        ...     date = library.get_date(matches, default_date=default_date)
        ...     ds = ds.assign_coords(time=[date])
        ...     return ds
        ...
        ... ds = xr.open_mfdataset(finder.get_files(),
        ...                        preprocess=finder.get_func_process_filename(
        ...     process, default_date={'hour': 12}))
        """
        def f(ds):
            filename = ds.encoding['source']
            if relative:
                filename = os.path.relpath(filename, self.root)
            return func(ds, filename, self, *args, **kwargs)
        return f

    def set_pregex(self, pregex: str, **replacements: str):
        """Set pre-regex.

        Apply replacements.
        """
        pregex = pregex.strip()
        for k, z in replacements.items():
            pregex = pregex.replace("%({:s})".format(k), z)
        self.pregex = pregex

    def scan_pregex(self):
        """Scan pregex for matchers.

        Add matchers objects to self.
        Set segments attribute.
        """
        splits = [0]
        self.matchers = []
        for i, m in enumerate(re.finditer(Matcher.REGEX, self.pregex)):
            self.matchers.append(Matcher(m, i))
            splits += [m.start(), m.end()]
        self.segments = [self.pregex[i:j]
                         for i, j in zip(splits, splits[1:]+[None])]

        # Replace matcher by its regex
        for idx, m in enumerate(self.matchers):
            self.segments[2*idx+1] = '({})'.format(m.get_regex())

    def update_regex(self):
        """Update regex.

        Set fixed matchers. Re-compile pattern. Scrap previous scanning.
        """
        segments = self.segments.copy()
        if not self.use_regex:
            for i, s in enumerate(segments):
                # Escape outside matchers
                segments[i] = s if i % 2 == 1 else re.escape(s)

        for idx, value in self.fixed_matchers.items():
            if not isinstance(value, (list, tuple)):
                value = [value]
            value = [v if isinstance(v, str)
                     else re.escape(self.matchers[idx].format(v))
                     for v in value]
            segments[2*idx+1] = '({})'.format('|'.join(value))

        self.regex = ''.join(segments)
        self.pattern = re.compile(self.regex)
        self.scanned = False
        self._files = []

    def find_files(self):
        """Find files to scan.

        Sort files alphabetically.

        Raises
        ------
        AttributeError
            If no regex is set.
        IndexError
            If no files are found in the filetree.
        """
        if self.regex is None:
            self.create_regex()
        if self.regex == '':
            raise AttributeError("Finder is missing a regex.")

        subpatterns = [re.compile(rgx)
                       for rgx in self.regex.split(os.path.sep)]
        files = []
        for dirpath, dirnames, filenames in os.walk(self.root):
            # Feels hacky, better way ?
            depth = dirpath.count(os.sep) - self.root.count(os.sep)
            pattern = subpatterns[depth]

            if depth == len(subpatterns)-1:
                dirnames.clear()  # Look no deeper
                files += [os.path.relpath(os.path.join(dirpath, f), self.root)
                          for f in filenames]
            else:
                dirlogs = dirnames[:3]
                if len(dirnames) > 3:
                    dirlogs += ['...']
                log.debug('depth: %d, pattern: %s, folders:\n\t%s',
                          depth, pattern.pattern, '\n\t'.join(dirlogs))

            # Removes directories not matching regex
            # We do double regex on directories, good enough
            to_remove = [d for d in dirnames if not pattern.fullmatch(d)]
            for d in to_remove:
                dirnames.remove(d)

        files.sort()

        if len(files) == 0:
            raise IndexError(f"No files were found in {self.root}")

        files_matched = []
        for f in files:
            try:
                matches = self.get_matches(f, relative=True)
            except ValueError:  # Filename did not match pattern
                pass
            else:
                files_matched.append((f, matches))

        filelogs = files[:3]
        if len(files) > 3:
            filelogs += ['...']
        log.debug("regex: %s, files:\n\t%s", self.regex, '\n\t'.join(filelogs))
        log.debug("Found %s matching files in %s",
                  len(files_matched), self.root)

        self.scanned = True
        self._files = files_matched

    def get_matchers(self, key: Union[int, str]) -> List[Matcher]:
        """Return list of matchers corresponding to key.

        Parameters
        ----------
        key: int, str, or list of int
            Can be matcher index, name, or group+name combination with the
            syntax: 'group:name'.

        Returns
        -------
        List of matchers corresponding to key.

        Raises
        ------
        KeyError: No matcher found.
        TypeError: Key type is not valid.
        """
        selected = get_matchers_indices(self.matchers, key)
        matchers = [self.matchers[i] for i in selected]
        return matchers



"""Matches management."""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

import re
import logging
from typing import Any, List, Union

from .format import Format


log = logging.getLogger(__name__)


class Matcher():
    """Manage a matcher inside the pre-regex.

    Parameters
    ----------
    m: re.Match
        Match object obtained to find matchers in the pre-regex.
    idx: int
        Index inside the pre-regex.

    Attributes
    ----------
    idx: int
        Index inside the pre-regex.
    group: str
        Group name.
    name: str
        Matcher name.
    rgx: str
        Regex.
    discard: bool
        If the matcher should not be used when retrieving values from matches.
    fmt: :class:`Format<filefinder.format.Format>`
        Format string object.
    match: str
        The string that created the matcher `%(match)`.
    """

    DEFAULT_ELTS = {
        "I": [r"\d+", 'd'],
        "Y": [r"\d{4}", '04d'],
        "m": [r"\d\d", '02d'],
        "d": [r"\d\d", '02d'],
        "j": [r"\d{3}", '03d'],
        "H": [r"\d\d", '02d'],
        "M": [r"\d\d", '02d'],
        "S": [r"\d\d", '02d'],
        "x": [r"%Y%m%d", '08d'],
        "X": [r"%H%M%S", '06d'],
        "F": [r"%Y-%m-%d", 's'],
        "B": [r"[a-zA-Z]*", 's'],
        "text": [r'\w', 's'],
        "char": [r"\S*", 's']
    }
    """Regex str for each type of element."""

    REGEX = (r"%\((?:(?P<group>\w*):)?"
             r"(?P<name>\w*)"
             r"(:fmt=(?P<fmt>.*?))?"
             r"(:rgx=(?P<rgx>.*?))?"
             r"(?P<discard>:discard)?\)")
    """Regex to find matcher in pre-regex."""

    def __init__(self, m: re.Match, idx: int = 0):
        self.idx = idx
        self.group = None
        self.name = None
        self.rgx = None
        self.discard = False
        self.fmt = None
        self.match = ''

        self.set_matcher(m)

    def __repr__(self):
        return '\n'.join([super().__repr__(), self.__str__()])

    def __str__(self):
        s = ''
        if self.group:
            s += self.group + ':'
        s += '{}:{:d}'.format(self.name, self.idx)
        return s

    def set_matcher(self, m: re.Match):
        """Find attributes from match.

        Raises
        ------
        NameError
            No name.
        ValueError
            Empty custom regex.
        """
        self.match = m.group()[2:-1]  # slicing removes %()
        self.group = m.group('group')
        self.name = m.group('name')
        self.discard = m.group('discard') is not None

        rgx = m.group('rgx')
        fmt = m.group('fmt')

        if self.name is None:
            raise NameError("Matcher name cannot be empty.")
        if rgx is not None and rgx == '':
            raise ValueError("Matcher custom regex cannot be empty.")
        if fmt is not None and fmt == '':
            raise ValueError("Matcher custom format cannot be empty.")

        # Set defaults if name is known
        if self.name in self.DEFAULT_ELTS:
            self.rgx, fmt_def = self.DEFAULT_ELTS[self.name]
            self.fmt = Format(fmt_def)

        # Override default format
        if fmt:
            self.fmt = Format(fmt)
            if not rgx:  # No need to generate rgx if it is provided
                self.rgx = self.fmt.generate_expression()

        # Override regex
        if rgx:
            self.rgx = rgx

    def format(self, value: Any):
        return self.fmt.format(value)

    def get_regex(self) -> str:
        """Get matcher regex.

        Replace the matchers name by regex from `Matcher.NAME_RGX`. If there is
        a custom regex, recursively replace '%' followed by a single letter by
        the corresponding regex from `NAME_RGX`. '%%' is replaced by a single
        percentage character.

        Raises
        ------
        KeyError
            Unknown replacement.
        """
        def replace(match):
            group = match.group(1)
            if group == '%':
                return '%'
            if group in self.DEFAULT_ELTS:
                replacement = self.DEFAULT_ELTS[group][0]
                if '%' in replacement:
                    return self.get_regex(replacement)
                return replacement
            raise KeyError("Unknown replacement '{}'.".format(match.group(0)))

        return re.sub("%([a-zA-Z%])", replace, self.rgx)


class Match:
    """Match extract from a filename.

    Attributes
    ----------
    matcher: Matcher
        Matcher used to get this match.
    match_str: str
        String matched in the filename.
    start: int
        Start index of match in the filename.
    end: int
        End index of match in the filename.
    match_parsed: any
        Parsed value. None if parsing was not successful.
    """

    def __init__(self, matcher: Matcher, match: re.Match, group: int):
        self.matcher = matcher
        self.match_str = match.group(group+1)
        self.start = match.start(group+1)
        self.end = match.end(group+1)

        self.match_parsed = None
        if matcher.fmt is not None:
            try:
                self.match_parsed = matcher.fmt.parse(self.match_str)
            except Exception:
                log.warning('Failed to parse for matcher %s', str(matcher))

    def __repr__(self):
        return '\n'.join([super().__repr__(), self.__str__()])

    def __str__(self):
        return str(self.matcher) + ' = {}'.format(self.match_str)

    def get_match(self, parsed: bool = True):
        """Get match string or value.

        If `parsed` is true, and the parsing was successful, return the
        parsed value instead of the matched string.
        """
        if parsed and self.match_parsed is not None:
            return self.match_parsed
        return self.match_str


class Matches:
    """Store multiples matches.

    Attributes
    ----------
    matches: list of :class:`Match`
        Matches for a single filename.
    matchers: list of :class:`Matcher`
        Matchers used.

    Raises
    ------
    ValueError: Filename did not match pattern.
    IndexError: Not as many matches as matchers.
    """

    def __init__(self, matchers: List[Matcher], filename: str,
                 pattern: re.Pattern):
        self.matches = []
        self.matchers = matchers

        m = pattern.fullmatch(filename)
        if m is None:
            raise ValueError("Filename did not match pattern.")
        if len(m.groups()) != len(matchers):
            raise IndexError("Not as many matches as matchers.")

        for i in range(len(matchers)):
            self.matches.append(Match(matchers[i], m, i))

    def __repr__(self):
        return '\n'.join([super().__repr__(), self.__str__()])

    def __str__(self):
        return '\n'.join([str(m) for m in self.matches])

    def __getitem__(self, key: Union[int, str]):
        return self.get_matches(key)

    def __iter__(self):
        return iter(self.matches)

    def __len__(self):
        return len(self.matches)

    def get_matches(self, key: Union[int, str]) -> Union[Match, List[Match]]:
        """Get matches corresponding to key.

        See :func:`Finder.get_matchers
        <filefinder.finder.Finder.get_matchers>` for details on
        `key` argument.

        Returns
        -------
        List of Match corresponding to the key. If only one Match corresponds,
        return it directly.
        """
        selected = get_matchers_indices(self.matchers, key)
        matches = [self.matches[k] for k in selected]
        if len(matches) == 1:
            return matches[0]
        return matches


def get_matchers_indices(matchers: List[Matcher],
                         key: Union[int, str]) -> List[int]:
    """Get list of matchers indices corresponding to key.

    Key can be an integer index, or a string of the name, or a combination
    of the group and the name with the syntax 'group:name'

    Raises
    ------
    IndexError: No matcher found corresponding to the key
    TypeError: Key is not int or str
    """
    if isinstance(key, int):
        return [key]
    if isinstance(key, str):
        k = key.split(':')
        if len(k) == 1:
            name, group = k[0], None
        else:
            group, name = k[:2]
        selected = []
        for i, m in enumerate(matchers):
            if m.name == name and (group is None or group == m.group):
                selected.append(i)

        if len(selected) == 0:
            raise IndexError(f"No matcher found for key '{key}'")
        return selected

    raise TypeError("Key must be int or str.")



"""Generate regex from string format, and parse strings.

Parameters of the format-string are retrieved.
See `Format Mini Language Specification
<https://docs.python.org/3/library/string.html#formatspec>`__.

Thoses parameters are then used to generate a regular expression, or to parse
a string formed from the format.

Only 's', 'd', 'f', 'e' and 'E' formats types are supported.

The width of the format string is not respected when matching with a regular
expression.

The parsing is quite naive and can fail on some cases.
See :func:`Format.parse` for details.

The regex generation and parsing are tested in `tests/unit/test_format.py`.
"""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

import re
from typing import Any, Tuple, Union


def autoprop(*props):
    """Generate properties for class."""
    def factory_get(name):
        def getter(self):
            return self.params[name]
        return getter

    def factory_set(name):
        def setter(self, value):
            self.params[name] = value
        return setter

    def decorator(cls):
        for name in props:
            prop = property(factory_get(name), factory_set(name))
            setattr(cls, name, prop)
        return cls

    return decorator


@autoprop('fill', 'align', 'sign', 'alternate', 'zero',
          'width', 'grouping', 'precision', 'type')
class Format:
    """Parse a format string.

    Out of found parameters:
    - generate regular expression
    - format value
    - parse string into value

    Parameters
    ----------
    fmt: str
        Format string.
    """

    ALLOWED_TYPES = 'fdeEs'

    def __init__(self, fmt: str):
        self.fmt = fmt
        self.parse_params(fmt)
        self.set_defaults()

    def parse_params(self, format: str):
        """Parse format parameters."""
        p = (r"((?P<fill>.)?(?P<align>[<>=^]))?"
             r"(?P<sign>[-+ ])?(?P<alternate>#)?"
             r"(?P<zero>0)?(?P<width>\d+?)?"
             r"(?P<grouping>[,_])?"
             r"(?P<precision>\.\d+?)?"
             r"(?P<type>[a-zA-Z])")
        m = re.fullmatch(p, format)
        if m is None:
            raise ValueError("Format spec not valid.")
        self.params = m.groupdict()
        if not self.type or self.type not in self.ALLOWED_TYPES:
            raise ValueError('format spec %r not supported' % self.type)

    def set_defaults(self):
        """Set parameters defaults values."""
        if self.type in 'dfeE':
            defaults = dict(
                align='>',
                fill=' ',
                sign='-',
                width='0',
                precision='.6'
            )
            self.alternate = self.alternate == '#'
            self.zero = self.zero == '0'
            if self.align is None and self.zero:
                self.fill = '0'
            for k, v in defaults.items():
                if self.params[k] is None:
                    self.params[k] = v
            self.width = int(self.width)
            self.precision = int(self.precision[1:])

    def format(self, value: Any) -> str:
        """Return formatted string."""
        return '{{:{}}}'.format(self.fmt).format(value)

    def generate_expression(self) -> str:
        """Generate regex from format string."""
        if self.type == 'f':
            return self.generate_expression_f()
        if self.type == 'd':
            return self.generate_expression_d()
        if self.type == 's':
            return self.generate_expression_s()
        if self.type in 'eE':
            return self.generate_expression_e()

    def parse(self, s: str) -> Union[str, int, float]:
        """Parse string generated with format.

        This simply use int() and float() to parse strings. Those are thrown
        off when using fill characters (other than 0), or thousands groupings,
        so we remove these from the string.

        Parsing will fail when using the '-' fill character on a negative
        number, or when padding with numbers. If you use such formats, please
        contact me to explain me why in the hell you do.
        """
        if self.type == 'd':
            return self.parse_d(s)
        if self.type in 'feE':
            return self.parse_f(s)
        if self.type == 's':
            return s

    def generate_expression_s(self) -> str:
        return '.*?'

    def generate_expression_d(self) -> str:
        rgx = self.get_left_point()
        return self.insert_in_alignement(rgx)

    def generate_expression_f(self) -> str:
        rgx = self.get_left_point()
        rgx += self.get_right_point()
        return self.insert_in_alignement(rgx)

    def generate_expression_e(self) -> str:
        rgx = r'\d'
        rgx += self.get_right_point()
        rgx += r'{}[+-]\d+?\d'.format(self.type)
        return self.insert_in_alignement(rgx)

    def insert_in_alignement(self, rgx: str) -> str:
        align, loc = self.get_align()
        out_rgx = ''

        if loc in ['left', 'center']:
            out_rgx += align

        out_rgx += self.get_sign()

        if loc == 'middle':
            out_rgx += align

        out_rgx += rgx

        if loc in ['right', 'center']:
            out_rgx += align

        return out_rgx

    def get_sign(self) -> str:
        """Get sign regex."""
        if self.sign == '-':
            rgx = '-?'
        elif self.sign == '+':
            rgx = r'[+-]'
        elif self.sign == ' ':
            rgx = r'[\s-]'
        else:
            raise KeyError("Sign not in {+- }")
        return rgx

    def get_align(self) -> Tuple[str, str]:
        """Get alignment with fill regex and its location."""
        rgx = ''
        if self.width > 0:
            rgx += '{}*'.format(re.escape(self.fill))

        loc = {
            '=': 'middle',
            '>': 'left',
            '<': 'right',
            '^': 'center'
        }[self.align]

        return rgx, loc

    def get_left_point(self) -> str:
        """Get regex for numbers left of decimal point."""
        if self.grouping is not None:
            rgx = r'\d?\d?\d(?:{}\d{{3}})*'.format(self.grouping)
        else:
            rgx = r'\d+'
        return rgx

    def get_right_point(self) -> str:
        rgx = ''
        if self.precision != 0 or self.alternate:
            rgx += r'\.'
        if self.precision != 0:
            rgx += r'\d{{{:d}}}'.format(self.precision)
        return rgx

    def parse_d(self, s: str) -> int:
        """Parse integer from formatted string. """
        return int(self.remove_special(s))

    def parse_f(self, s: str) -> float:
        """Parse float from formatted string."""
        return float(self.remove_special(s))

    def remove_special(self, s: str) -> int:
        """Remove special characters.

        Remove characters that throw off int() and float() parsing.
        Namely fill and grouping characters.
        Will remove fill, except when fill is zero (parsing functions are
        okay with that).
        """
        to_remove = [',', '_']  # Any grouping char
        if self.fill != '0':
            to_remove.append(re.escape(self.fill))
        pattern = '[{}]'.format(''.join(to_remove))
        return re.sub(pattern, '', s)



"""Functions to retrieve values from filename."""

# This file is part of the 'filefinder' project
# (http://github.com/Descanonge/filefinder) and subject
# to the MIT License as defined in the file 'LICENSE',
# at the root of this project. © 2021 Clément Haëck

import logging
from typing import Dict, List

from datetime import datetime, timedelta

log = logging.getLogger(__name__)


def get_date(matches: List, default_date: Dict = None,
             group: str = None) -> datetime:
    """Retrieve date from matched elements.

    If any element is not found in the filename, it will be replaced by the
    element in the default date. If no match is found, None is returned.

    Supports matches with names from `Matcher.NAME_RGX`.

    Parameters
    ----------
    matches: list
        Matches from a filename, returned by `Finder.get_matches`
    group: str
        If not None, restrict matcher to this group.
    default_date: dict, optional
        Default date. Dictionnary with keys: year, month, day, hour, minute,
        and second. Defaults to 1970-01-01 00:00:00

    Raises
    ------
    KeyError: If no matchers are found to create a date from.
    """
    date = {"year": 1970, "month": 1, "day": 1,
            "hour": 00, "minute": 0, "second": 0}

    if default_date is None:
        default_date = {}
    date.update(default_date)

    elts = {m.matcher.name: m.get_match() for m in matches
            if (not m.matcher.discard
                and (group is None or m.matcher.group == group))}

    elts_needed = {'x', 'X', 'Y', 'm', 'd', 'B', 'j', 'H', 'M', 'S', 'F'}
    if len(set(elts.keys()) & elts_needed) == 0:
        log.warning("No matchers to retrieve a date from."
                    " Returning default date.")

    elt = elts.pop("F", None)
    if elt is not None:
        elts["Y"] = elt[:4]
        elts["m"] = elt[5:7]
        elts["d"] = elt[8:10]

    elt = elts.pop("x", None)
    if elt is not None:
        elts["Y"] = elt[:4]
        elts["m"] = elt[4:6]
        elts["d"] = elt[6:8]

    elt = elts.pop("X", None)
    if elt is not None:
        elts["H"] = elt[:2]
        elts["M"] = elt[2:4]
        if len(elt) > 4:
            elts["S"] = elt[4:6]

    elt = elts.pop("Y", None)
    if elt is not None:
        date["year"] = int(elt)

    elt = elts.pop("m", None)
    if elt is not None:
        date["month"] = int(elt)

    elt = elts.pop("B", None)
    if elt is not None:
        elt = _find_month_number(elt)
        if elt is not None:
            date["month"] = elt

    elt = elts.pop("d", None)
    if elt is not None:
        date["day"] = int(elt)

    elt = elts.pop("j", None)
    if elt is not None:
        elt = datetime(date["year"], 1, 1) + timedelta(days=int(elt)-1)
        date["month"] = elt.month
        date["day"] = elt.day

    elt = elts.pop("H", None)
    if elt is not None:
        date["hour"] = int(elt)

    elt = elts.pop("M", None)
    if elt is not None:
        date["minute"] = int(elt)

    elt = elts.pop("S", None)
    if elt is not None:
        date["second"] = int(elt)

    return datetime(**date)


def _find_month_number(name: str) -> int:
    """Find a month number from its name.

    Name can be the full name (January) or its three letter abbreviation (jan).
    The casing does not matter.
    """
    names = ['january', 'february', 'march', 'april',
             'may', 'june', 'july', 'august', 'september',
             'october', 'november', 'december']
    names_abbr = [c[:3] for c in names]

    name = name.lower()
    if name in names:
        return names.index(name)
    if name in names_abbr:
        return names_abbr.index(name)

    return None




from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))


def get_long_description(rel_path):
    with open(path.join(here, rel_path)) as file:
        return file.read()


def get_version(rel_path):
    with open(path.join(here, rel_path)) as file:
        lines = file.read().splitlines()
    for line in lines:
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string")


CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'License :: OSI Approved :: MIT License',

    'Programming Language :: Python',
    'Programming Language :: Python :: 3 :: Only',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
    'Operating System :: OS Independent',

    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'Topic :: Utilities'
]


setup(name='filefinder',
      version=get_version('src/filefinder/__init__.py'),

      description="Find files using a simple syntax.",
      long_description=get_long_description('README.md'),
      long_description_content_type='text/markdown',

      keywords='find files filename regular expression regex xarray',
      classifiers=CLASSIFIERS,

      url='https://github.com/Descanonge/filefinder',
      project_urls={
          'Source': 'https://github.com/Descanonge/filefinder',
          'Documentation': 'https://filefinder.readthedocs.io'
      },

      author='Clément Haëck',
      author_email='clement.haeck@posteo.net',

      python_requires='>=3.7',

      package_dir={'': 'src'},
      packages=find_packages(where='src'),
      )
